(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2005-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: write_typed_value.ml,v 1.24 2017/11/06 09:12:17 fclement Exp $ *)

module Nat = Unatural;;
module S = Ustring;;
module A = Uarray;;
module M = Umatrix;;
module D = Udebug;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _10_ = Nat.unsafe_of_int 10
and _31_ = Nat.unsafe_of_int 31
and _34_ = Nat.unsafe_of_int 34
and _92_ = Nat.unsafe_of_int 92
and _100_ = Nat.unsafe_of_int 100
and _127_ = Nat.unsafe_of_int 127
;;

module Op = Options_pio;;

open Layout;;
open Typed_value;;

let qual = Upervasives.qualify "Write_typed_value";;

let debug fun_name = Op.pf.D.debug (Some (qual fun_name));;

(* For additional information *)

let w_value_type ob t =
  debug "w_value_type" "start writing";
  Printf.bprintf ob "%a" Value_type.to_ob t;
  debug "w_value_type" "finish writing";;

let w_layout ob l =
  Printf.bprintf ob "%a" Layout.to_ob l
;;

let print_natural ob nat = Printf.bprintf ob "%i" (Nat.to_int nat);;

let w_size ob = Printf.bprintf ob "<%a>" print_natural
and w_sizes ob =  Printf.bprintf ob "<%a>" (fun ob nats ->
  let rec loop ob = function
    | [] -> ()
    | [nat] -> print_natural ob nat
    | nat :: nats -> Printf.bprintf ob "%a, %a" print_natural nat loop nats in
  loop ob nats)
;;

(* The particular case for matrices (2-d arrays). *)
let w_size_matrix ob (n1, n2) = w_sizes ob [n1; n2];;

(* w_tuple *)
let w_couple w_c ob (x1, x2) =
  Printf.bprintf ob "(%a, %a)" w_c x1 w_c x2
and w_triple w_c ob (x1, x2, x3) =
  Printf.bprintf ob "(%a, %a, %a)" w_c x1 w_c x2 w_c x3
;;

(* w_simple_lexem *)
let string_to_ascii ob s =
  Nat.for_loop _0_ (Nat.pred (S.length s))
    (fun i ->
      let c = Nat.of_char (S.unsafe_get s i) in
      if c > _31_ && c < _127_ then
        Buffer.add_char ob (Nat.to_char c) else begin
      Buffer.add_char ob '\\';
      if c = _92_ (* backslash *) || c = _34_ (* double quote *) then
        Buffer.add_char ob (Nat.to_char c) else begin
      if c < _10_ then Buffer.add_char ob '0';
      if c < _100_ then Buffer.add_char ob '0';
      Buffer.add_string ob (Nat.to_string c) end
      end)
;;

let output_ascii_string ob s =
  Buffer.add_char ob '\"';
  string_to_ascii ob s;
  Buffer.add_char ob '\"';
;;

let w_bool ob = Printf.bprintf ob "%B"

and w_string ob s =
  let len = S.length s in
  Printf.bprintf ob "%a\n%a" w_size len output_ascii_string s

and w_int ob = Printf.bprintf ob "%i"

and w_native_int ob = Printf.bprintf ob "%ni"

and w_int32 ob = Printf.bprintf ob "%li"

and w_int64 ob = Printf.bprintf ob "%Li"

and w_float ob f =
  if Op.get_binary_mode () then
    let s = Upervasives.bit_string_of_float f in
    Printf.bprintf ob "&<%i>%s" (String.length s) s
  else
    Printf.bprintf ob "% 23.16e" f
    (* The intent of this format is to align all floating point numbers on
       their decimal point. Thus we use the space flag to replace the absent
       sign for nonnegative values. *)
;;

(* w_tuple_lexem *)
let w_bool_couple = w_couple w_bool
and w_bool_triple = w_triple w_bool
and w_int_couple = w_couple w_int
and w_int_triple = w_triple w_int
and w_native_int_couple = w_couple w_native_int
and w_native_int_triple = w_triple w_native_int
and w_int32_couple = w_couple w_int32
and w_int32_triple = w_triple w_int32
and w_int64_couple = w_couple w_int64
and w_int64_triple = w_triple w_int64
and w_float_couple = w_couple w_float
and w_float_triple = w_triple w_float
;;

(* Writing items *)
let w_item w_lexem ob lexem = Printf.bprintf ob "%a;\n" w_lexem lexem;;

let w_items w_lexem ob v = A.iter (w_item w_lexem ob) v;;

(* For scalars *)

let w_scalar ob = function
    (* %B *)
  | Sbool b -> w_bool ob b
    (* %S *)
  | Sstring s -> w_string ob s
    (* %i *)
  | Sint i -> w_int ob i
    (* %ni *)
  | Snative_int i -> w_native_int ob i
    (* %li *)
  | Sint32 i -> w_int32 ob i
    (* %Li *)
  | Sint64 i -> w_int64 ob i
    (* %f or %bf according to the binary mode flag. *)
  | Sfloat f -> w_float ob f
    (* Couples *)
  | Sbool_couple bc -> w_bool_couple ob bc
  | Sint_couple ic -> w_int_couple ob ic
  | Snative_int_couple ic -> w_native_int_couple ob ic
  | Sint32_couple ic -> w_int32_couple ob ic
  | Sint64_couple ic -> w_int64_couple ob ic
  | Sfloat_couple fc -> w_float_couple ob fc
    (* Triples *)
  | Sbool_triple bt -> w_bool_triple ob bt
  | Sint_triple it -> w_int_triple ob it
  | Snative_int_triple it -> w_native_int_triple ob it
  | Sint32_triple it -> w_int32_triple ob it
  | Sint64_triple it -> w_int64_triple ob it
  | Sfloat_triple ft -> w_float_triple ob ft
;;

(* For vectors *)

let w_vector w_lexem ob v =
  let len = A.length v in
  Printf.bprintf ob "[1\n%a\n%a1]" w_size len (w_items w_lexem) v
;;

let w_vect ob = function
  | Vbool bv -> w_vector w_bool ob bv
  | Vbool_couple bcv -> w_vector w_bool_couple ob bcv
  | Vbool_triple btv -> w_vector w_bool_triple ob btv
  | Vstring sv -> w_vector w_string ob sv
  | Vint iv -> w_vector w_int ob iv
  | Vint_couple icv -> w_vector w_int_couple ob icv
  | Vint_triple itv -> w_vector w_int_triple ob itv
  | Vnative_int iv -> w_vector w_native_int ob iv
  | Vnative_int_couple icv -> w_vector w_native_int_couple ob icv
  | Vnative_int_triple itv -> w_vector w_native_int_triple ob itv
  | Vint32 iv -> w_vector w_int32 ob iv
  | Vint32_couple icv -> w_vector w_int32_couple ob icv
  | Vint32_triple itv -> w_vector w_int32_triple ob itv
  | Vint64 iv -> w_vector w_int64 ob iv
  | Vint64_couple icv -> w_vector w_int64_couple ob icv
  | Vint64_triple itv -> w_vector w_int64_triple ob itv
  | Vfloat fv -> w_vector w_float ob fv
  | Vfloat_couple fcv -> w_vector w_float_couple ob fcv
  | Vfloat_triple ftv -> w_vector w_float_triple ob ftv
;;

(* For matrices *)

let w_line w_lexem ob l =
  Printf.bprintf ob "[|\n%a|]" (w_items w_lexem) l
;;

let w_line_item w_lexem ob = w_item (w_line w_lexem) ob
;;

let pcl1 = (_0_, [ _1_; ])
and pfl1 = (_1_, [ _0_; ])
;;

let w_line_items layout w_lexem ob m =
  match layout with
  | C -> M.iter_i (w_line_item w_lexem ob) m
  | F -> M.iter_j (w_line_item w_lexem ob) m
  | Permutation p when p = pcl1 -> M.iter_i (w_line_item w_lexem ob) m
  | Permutation p when p = pfl1 -> M.iter_j (w_line_item w_lexem ob) m
  | Permutation _ -> assert false
;;

let w_matrix l w_lexem ob m =
  Printf.bprintf ob
    "[2\n%a\n%a\n%a2]"
    w_size_matrix (M.size m)
    w_layout l
    (w_line_items l w_lexem) m
;;

let w_mat ob = function
  | Mcint rim -> w_matrix C w_int ob rim
  | Mfint cim -> w_matrix F w_int ob cim
  | Mcfloat rfm -> w_matrix C w_float ob rfm
  | Mffloat cfm -> w_matrix F w_float ob cfm
;;

let w_value ob = function
  | Scalar s -> w_item w_scalar ob s
  | Vector v -> w_item w_vect ob v
  | Matrix m -> w_item w_mat ob m
;;

let write_typed_value ob v =
  let t = type_of_value v in
  Printf.bprintf ob "begin\n%a\n%aend\n\n" w_value_type t w_value v
;;
