(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rdec2.ml,v 1.17 2017/01/04 11:58:30 fclement Exp $ *)

(* 2D regular rectangular grids. *)

module Nat = Unatural;;
module Perv = Upervasives;;
module L = Ulist;;
module A = Uarray;;

let _0_ = Nat._0_
and _2_ = Nat._2_
;;


let a_get_first a =
  assert (A.length a > _0_);
  (* Now, we have A.length > 0.
     Therefore, the next A.unsafe_get will be safe. *)
  A.unsafe_get a _0_

and a_set_first a x =
  assert (A.length a > _0_);
  (* Now, we have A.length > 0.
     Therefore, the next A.unsafe_set will be safe. *)
  A.unsafe_set a _0_ x

and a_get_last a =
  let na = A.length a in
  assert (na > _0_);
  (* Now, we have na = A.length > 0.
     Therefore, the next Nat.pred will not fail and the next A.unsafe_get
     will be safe. *)
  A.unsafe_get a (Nat.pred na)

and a_set_last a x =
  let na = A.length a in
  assert (na > _0_);
  (* Now, we have na = A.length > 0.
     Therefore, the next Nat.pred will not fail and the next A.unsafe_set
     will be safe. *)
  A.unsafe_set a (Nat.pred na) x
;;

(* Regular decomposition. *)

type length = Perv.length
and index = Perv.index
;;

type size = length * length;;

type rectangle =
  {
   x0 : float;
   y0 : float;
   lx : float;
   ly : float;
  }

and rectangular_grid =
  {
   domain : rectangle;
   nx : length;
   ny : length;
  }

and computational_rectangular_grid =
  {
   mesh : rectangular_grid;
   px : length;
   py : length;
  }

and regular_decomposition =
  {
   comp_grid : computational_rectangular_grid;
   nsbdx : length;
   nsbdy : length;
  }

and t = regular_decomposition
;;

let rectangle_of x0 y0 lx ly = { x0; y0; lx; ly; };;

let rectangular_grid_of x0 y0 lx ly nx ny =
  let r = rectangle_of x0 y0 lx ly in
  { domain = r; nx; ny; }
;;

let computational_rectangular_grid_of x0 y0 lx ly nx ny px py =
  let g = rectangular_grid_of x0 y0 lx ly nx ny in
  { mesh = g; px; py; }
;;

let number_of_subdomains_of_regular_decomposition dec =
  (dec.nsbdx, dec.nsbdy)
;;

let print_rectangle oc r =
  Printf.fprintf oc "%f %f\n%f %f\n" r.x0 r.y0 r.lx r.ly
;;

let print_rectangular_grid oc g =
  Printf.fprintf oc "%a%d %d\n"
    print_rectangle g.domain (Nat.to_int g.nx) (Nat.to_int g.ny)
;;

let print_computational_rectangular_grid oc g =
  Printf.fprintf oc "%a%d %d\n"
    print_rectangular_grid g.mesh (Nat.to_int g.px) (Nat.to_int g.py)
;;

let read_rectangle ib =
  Scanf.bscanf ib "x0 = %f; y0 = %f;\nlx = %f; ly = %f;\n" rectangle_of
;;

let read_rectangular_grid ib =
  let r = read_rectangle ib in
  Scanf.bscanf ib "nx = %d; ny = %d;\n"
    (fun nx ny ->
      let nx = Nat.of_int nx
      and ny = Nat.of_int ny in
      { domain = r; nx; ny; })
;;

let read_computational_rectangular_grid ib =
  let g = read_rectangular_grid ib in
  Scanf.bscanf ib "px = %d; py = %d;\n"
    (fun px py ->
      let px = Nat.of_int px
      and py = Nat.of_int py in
      { mesh = g; px; py; })
;;

let read_regular_decomposition ib =
  let g = read_computational_rectangular_grid ib in
  Scanf.bscanf ib "nsbdx = %d; nsbdy = %d;\n"
    (fun nsbdx nsbdy ->
      let nsbdx = Nat.of_int nsbdx
      and nsbdy = Nat.of_int nsbdy in
      { comp_grid = g; nsbdx; nsbdy; })
;;

exception Wrong_Input of string;;

let check_regular_decomposition
    {
     comp_grid = { mesh = { domain = { lx; ly; _; }; nx; ny; }; px; py; };
     nsbdx; nsbdy;
    } =
  if lx <= 0. || ly <= 0. then
    raise (Wrong_Input "Nonpositive length!") else
  if nx = _0_ || ny = _0_ then
    raise (Wrong_Input "Zero interval!") else
  if px = _0_ || py = _0_ then
    raise (Wrong_Input "Zero overlapping!") else
  if px > Nat.div nx nsbdx || py > Nat.div ny nsbdy then
    raise (Wrong_Input "Overlapping is too large!") else
  if nsbdx = _0_ || nsbdy = _0_ then
    raise (Wrong_Input "Zero subdomain!") else
  if nsbdx > nx || nsbdy > ny then
    raise (Wrong_Input "Too many subdomains!") else
  if Nat.rem nx nsbdx <> _0_ || Nat.rem ny nsbdy <> _0_ then
    raise (Wrong_Input
      "Number of subdomains does not divide number of intervals!")
;;

let parse_rdec_config_file fname =
  let ib =
    try Scanf.Scanning.from_file fname with
    | Sys_error s -> prerr_endline s; exit 6 in
  let rdec = read_regular_decomposition ib in
  Scanf.Scanning.close_in ib;
  rdec
;;

(* Subdomain. *)

type subdomain_index = index

and subdomain =
  {
   dec : t;
   number : subdomain_index;
  }
;;

type location =
  | EastSide
  | NorthSide
  | SouthSide
  | WestSide
;;

module OrderedLocation = struct
  type t = location;;
  let compare = Pervasives.compare;;
end
;;
module Position = Set.Make (OrderedLocation);;

exception One_d_index_out_of_range of index;;
exception Two_d_index_out_of_range of index * index;;

let two_indices_numbering nx ny i =
  if i >= Nat.mul nx ny then raise (One_d_index_out_of_range i) else
  (Nat.rem i nx, Nat.div i nx)
;;

let one_index_numbering nx ny (ix, iy) =
  if ix >= nx || iy >= ny then raise (Two_d_index_out_of_range (ix, iy)) else
  Nat.add ix (Nat.mul iy nx)
;;

let rectangular_grid_of_subdomain { dec; number = i; } =
  let {
    comp_grid = { mesh = { domain = { x0; y0; lx; ly; }; nx; ny; }; _; };
    nsbdx; nsbdy;
  } = dec in
  let (ix, iy) = two_indices_numbering nsbdx nsbdy i in
  let lxi = lx /. Nat.to_float nsbdx
  and lyi = ly /. Nat.to_float nsbdy in
  let x0i = x0 +. Nat.to_float ix *. lxi
  and y0i = y0 +. Nat.to_float iy *. lyi in
  rectangular_grid_of x0i y0i lxi lyi (Nat.div nx nsbdx) (Nat.div ny nsbdy)
;;

let position_of_subdomain { dec = { nsbdx; nsbdy; _; }; number = i; } =
  (* The argument regular decomposition is assumed to be checked.
     In particular, we have nsbdx, nsbdy > 0.
     Therefore, the next Nat.pred will not fail. *)
  let (ix, iy) = two_indices_numbering nsbdx nsbdy i in
  let locations_to_add =
    [ (ix = _0_, WestSide);
      (ix = Nat.pred nsbdx, EastSide);
      (iy = _0_, SouthSide);
      (iy = Nat.pred nsbdy, NorthSide); ] in
  let rec loop accu = function
    | [] -> accu
    | (b, e) :: l ->
      if b then loop (Position.add e accu) l else
      loop accu l in
  loop Position.empty locations_to_add
;;

let complement pos =
  let fullset =
    Position.add EastSide
      (Position.add NorthSide
        (Position.add SouthSide
          (Position.add WestSide
            Position.empty))) in
  Position.diff fullset pos
;;

(* FIXME: prove safety of the next Nat.pred calls. *)

let neighbors_of ({ dec; number = i; } as subd) =
  let (ix, iy) = two_indices_numbering dec.nsbdx dec.nsbdy i in
  let boundaries = complement (position_of_subdomain subd) in
  let add_neighbor f l = function
    | EastSide -> f (Nat.succ ix, iy) :: l
    | NorthSide -> f (ix, Nat.succ iy) :: l
    | SouthSide -> f (ix, Nat.pred iy) :: l
    | WestSide -> f (Nat.pred ix, iy) :: l in
  let f = one_index_numbering dec.nsbdx dec.nsbdy in
  L.fold_left (add_neighbor f) [] (Position.elements boundaries)
;;

let neighbor_orientation_of_id { dec; number = i; } j =
  let (ix, iy) = two_indices_numbering dec.nsbdx dec.nsbdy i in
  let (jx, jy) = two_indices_numbering dec.nsbdx dec.nsbdy j in
  if Nat.succ ix = jx && iy = jy then Some EastSide else
  if ix = jx && Nat.succ iy = jy then Some NorthSide else
  if ix = jx && Nat.pred iy = jy then Some SouthSide else
  if Nat.pred ix = jx && iy = jy then Some WestSide else
  None
;;

let east_extension_of_computational_rectangular_grid
    { mesh = { domain = { x0; y0; lx; ly; }; nx; ny; }; px; py; } =
  let dx = lx /. Nat.to_float nx in
  computational_rectangular_grid_of
    x0 y0 (lx +. Nat.to_float px *. dx) ly
    (Nat.add nx px) ny px py

and north_extension_of_computational_rectangular_grid
    { mesh = { domain = { x0; y0; lx; ly; }; nx; ny; }; px; py; } =
  let dy = ly /. Nat.to_float ny in
  computational_rectangular_grid_of
    x0 y0 lx (ly +. Nat.to_float py *. dy)
    nx (Nat.add ny py) px py

and south_extension_of_computational_rectangular_grid
    { mesh = { domain = { x0; y0; lx; ly; }; nx; ny; }; px; py; } =
  let dy = ly /. Nat.to_float ny in
  computational_rectangular_grid_of
    x0 (y0 -. Nat.to_float py *. dy) lx (ly +. Nat.to_float py *. dy)
    nx (Nat.add ny py) px py

and west_extension_of_computational_rectangular_grid
    { mesh = { domain = { x0; y0; lx; ly; }; nx; ny; }; px; py; } =
  let dx = lx /. Nat.to_float nx in
  computational_rectangular_grid_of
    (x0 -. Nat.to_float px *. dx) y0 (lx +. Nat.to_float px *. dx) ly
    (Nat.add nx px) ny px py
;;

let computational_rectangular_grid_of_subdomain ({ dec; _; } as subd) =
  let pos = position_of_subdomain subd in
  let g0 =
    { dec.comp_grid with mesh = rectangular_grid_of_subdomain subd; } in
  let ext_of_grid g = function
    | EastSide -> east_extension_of_computational_rectangular_grid g
    | NorthSide -> north_extension_of_computational_rectangular_grid g
    | SouthSide -> south_extension_of_computational_rectangular_grid g
    | WestSide -> west_extension_of_computational_rectangular_grid g in
  let rec loop accu = function
    | [] -> accu
    | cp :: l -> loop (ext_of_grid accu cp) l in
  loop g0 (Position.elements (complement pos))
;;

(* Connectivity table. *)

type neighbor = index

and neighborhood =
  {
   east_n : neighbor option;
   north_n : neighbor option;
   south_n : neighbor option;
   west_n : neighbor option;
  }

and connectivity_table = neighborhood array
;;

let change_east_neighbor nbh e = { nbh with east_n = Some e; }
and change_north_neighbor nbh n = { nbh with north_n = Some n; }
and change_south_neighbor nbh s = { nbh with south_n = Some s; }
and change_west_neighbor nbh w = { nbh with west_n = Some w; }
;;

let neighborhood_of_subdomain ({ dec; number = i; } as subd) =
  let pos = position_of_subdomain subd in
  let nsbdx = dec.nsbdx in
  let nbh0 =
    { west_n = None; east_n = None; south_n = None; north_n = None; } in
  let change_neighborhood nbh = function
    | EastSide -> change_east_neighbor nbh (Nat.succ i)
    | NorthSide -> change_north_neighbor nbh (Nat.add i nsbdx)
    | SouthSide -> change_south_neighbor nbh (Nat.sub i nsbdx)
    | WestSide -> change_west_neighbor nbh (Nat.pred i) in
  let rec loop accu = function
    | [] -> accu
    | cp :: l -> loop (change_neighborhood accu cp) l in
  loop nbh0 (Position.elements (complement pos))
;;

let connectivity_table_of_regular_decomposition dec =
  let n = Nat.mul dec.nsbdx dec.nsbdy in
  A.init n (fun i -> neighborhood_of_subdomain { dec; number = i; })
;;

let east_neighbor tbl i = (A.get tbl i).east_n
and north_neighbor tbl i = (A.get tbl i).north_n
and south_neighbor tbl i = (A.get tbl i).south_n
and west_neighbor tbl i = (A.get tbl i).west_n
;;

let print_neighbor oc = function
  | None -> Printf.fprintf oc "None"
  | Some n -> Printf.fprintf oc "Some %d" (Nat.to_int n)
;;

let print_neighborhood oc nh =
  Printf.fprintf oc
    "{ west_n = %a; east_n = %a;\
      south_n = %a; north_n = %a; }"
    print_neighbor nh.east_n
    print_neighbor nh.north_n
    print_neighbor nh.south_n
    print_neighbor nh.west_n
;;

let print_connectivity_table oc tbl =
  let f i nh =
    Printf.fprintf oc "Subd %d: %a\n"
      (Nat.to_int i) print_neighborhood nh in
  A.iter_i f tbl
;;

(* Interface. *)

type edge_values = float array

and boundary =
  {
   east_ev : edge_values;
   north_ev : edge_values;
   south_ev : edge_values;
   west_ev : edge_values;
  }

and boundary_size =
  {
   size_x : Nat.t;
   size_y : Nat.t;
  }

and interface = boundary array
;;

let size_of_boundary bd =
  let size_x = A.length bd.east_ev
  and size_y = A.length bd.north_ev in
  assert (size_x = A.length bd.west_ev);
  assert (size_y = A.length bd.south_ev);
  { size_x; size_y; }
;;

let print_edge_values oc ev =
  let f v = Printf.fprintf oc "%f\n" v in
  A.iter f ev
;;

let print_boundary oc bd =
  Printf.fprintf oc "%a%a%a%a"
    print_edge_values bd.east_ev
    print_edge_values bd.north_ev
    print_edge_values bd.south_ev
    print_edge_values bd.west_ev
;;

let read_edge_values ib sz =
  let f _i = Scanf.bscanf ib "%f\n" Perv.id in
  A.init sz f
;;

let read_boundary ib { size_x; size_y; } =
  let east_ev = read_edge_values ib size_x
  and north_ev = read_edge_values ib size_y
  and south_ev = read_edge_values ib size_y
  and west_ev = read_edge_values ib size_x in
  { east_ev; north_ev; south_ev; west_ev; }
;;

let make_zero_edge_values n = A.make n 0.;;

let make_zero_boundary { size_x; size_y; } =
  let east_ev = make_zero_edge_values size_x
  and north_ev = make_zero_edge_values size_y
  and south_ev = make_zero_edge_values size_y
  and west_ev = make_zero_edge_values size_x in
  { east_ev; north_ev; south_ev; west_ev; }
;;

let make_zero_interface dec =
  let f i =
    let { mesh = { nx; ny; _; }; _; } =
      computational_rectangular_grid_of_subdomain { dec; number = i; } in
    let size_x = Nat.succ ny
    and size_y = Nat.succ nx in
    make_zero_boundary { size_x; size_y; } in
  A.init (Nat.mul dec.nsbdx dec.nsbdy) f
;;

let proj_boundary sd b1 (id, b2) =
  let copy = A.copy in
  let len = A.length in
  match neighbor_orientation_of_id sd id with
  | None -> invalid_arg "proj_boundary"
  | Some EastSide ->
    assert (len b1.east_ev = len b2.west_ev);
    { b1 with east_ev = copy b2.west_ev; }
  | Some NorthSide ->
    assert (len b1.north_ev = len b2.south_ev);
    { b1 with north_ev = copy b2.south_ev; }
  | Some SouthSide ->
    assert (len b1.south_ev = len b2.north_ev);
    { b1 with south_ev = copy b2.north_ev; }
  | Some WestSide ->
    assert (len b1.west_ev = len b2.east_ev);
    { b1 with west_ev = copy b2.east_ev; }
;;

let normalize_boundary ({ north_ev; south_ev; east_ev; west_ev } as b) =
  let le = A.length east_ev
  and ln = A.length north_ev
  and ls = A.length south_ev
  and lw = A.length west_ev in
  let mean x y = (x +. y) *. 0.5 in
  let ne =
    mean
      (A.get north_ev (Nat.pred ln))
      (A.get east_ev (Nat.pred le))
  and nw =
    mean
      (A.get north_ev _0_)
      (A.get west_ev (Nat.pred lw))
  and se =
    mean
      (A.get south_ev (Nat.pred ls))
      (A.get east_ev _0_)
  and sw =
    mean
      (A.get south_ev _0_)
      (A.get west_ev _0_) in
  A.set east_ev (Nat.pred le) ne;
  A.set east_ev _0_ se;
  A.set north_ev _0_ nw;
  A.set north_ev (Nat.pred ln) ne;
  A.set south_ev _0_ sw;
  A.set south_ev (Nat.pred ls) se;
  A.set west_ev _0_ sw;
  A.set west_ev (Nat.pred lw) nw;
  b
;;

let projection tbl intf =
  let f i =
    let je = east_neighbor tbl i
    and jn = north_neighbor tbl i
    and js = south_neighbor tbl i
    and jw = west_neighbor tbl i in
    let e =
      match je with
      | None -> A.make (A.length (A.get intf i).east_ev) 0.
      | Some je -> A.copy (A.get intf je).west_ev
    and n =
      match jn with
      | None -> A.make (A.length (A.get intf i).north_ev) 0.
      | Some jn -> A.copy (A.get intf jn).south_ev
    and s =
      match js with
      | None -> A.make (A.length (A.get intf i).south_ev) 0.
      | Some js -> A.copy (A.get intf js).north_ev
    and w =
      match jw with
      | None -> A.make (A.length (A.get intf i).west_ev) 0.
      | Some jw -> A.copy (A.get intf jw).east_ev in
    let ne = (a_get_last n +. a_get_last e) *. 0.5
    and nw = (a_get_first n +. a_get_last w) *. 0.5
    and se = (a_get_last s +. a_get_first e) *. 0.5
    and sw = (a_get_first s +. a_get_first w) *. 0.5 in
    a_set_first e se;
    a_set_first n nw;
    a_set_first s sw;
    a_set_first w sw;
    a_set_last e ne;
    a_set_last n ne;
    a_set_last s se;
    a_set_last w nw;
    {
     east_ev = A.copy e;
     north_ev = A.copy n;
     south_ev = A.copy s;
     west_ev = A.copy w;
    } in
  A.init (A.length tbl) f
;;

let for_all_subdomains (nsbdx, nsbdy) to_do =
  (* We assume that nsbdx, nsbdy > 0.
     Therefore, the next Nat.pred will not fail. *)
  let pnsbdx = Nat.pred nsbdx
  and pnsbdy = Nat.pred nsbdy in
  let num = ref (Nat.mul nsbdx (Nat.succ nsbdy)) in
  Nat.for_loop _0_ pnsbdy
    (fun _iy ->
      (* Loop invariant:
           !num = nsbdx * nsbdy - (iy - 1) * nsbdx,
           0 <= iy <= nsbdy - 1.
         Thus, we have 2 * nsbdx <= !num <= nsbdx * nsbdy + nsbdx.
         Therefore, the next Nat.sub will not fail. *)
      num := Nat.sub !num (Nat.mul _2_ nsbdx);
      Nat.for_loop _0_ pnsbdx
        (fun _ix -> to_do !num; Nat.incr num))
;;
