(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: skl.ml,v 1.42 2017/05/15 13:58:09 fclement Exp $ *)

(* Skeletons operations. *)

module Nat = Unatural;;
module Perv = Upervasives;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
;;

let qual = Perv.qualify "Skl";;
let invalid_arg = Perv.invalid_arg_qual qual;;

type ('a, 'b) sum = ('a, 'b) Sk.sum =
  | Inl of 'a
  | Inr of 'b
;;

type ('a, 'b) t = ('a, 'b) Sk.Skl.t
and ('a, 'b) skel = ('a, 'b) t
;;

let default_color = 1;;

let atom (params, llexp) =
  Sk.c_atom params llexp default_color
;;

let to_sarray_gen = fun () -> Sklstream.to_sarray;;
let of_sarray_gen = fun () -> Sklstream.of_sarray;;

Expor.register to_sarray_gen;;
Expor.register of_sarray_gen;;

let get_exported f =
  match Expor.is_registered f with
  | None -> assert false
  | Some x -> atom ((), x)
;;

(* Creating skeletons. *)

let id = Sk.c_id;;

let make_skl w =
  let ww x = Sklstream.map (w x) in
  let rc = Expor.register ww in
  fun x -> atom (x, rc)
;;

(* Composing skeletons. *)

let loop (p, skl) = Sk.c_loop p skl;;

let pipeline = Sk.c_pipe;;

let ( ||| ) = pipeline;;

let product = Sk.c_prod;;

let ( *** ) = product;;

let sum (skl1, skl2) = Sk.c_sum skl1 skl2;;

let ( +++ ) a b = sum (a, b);;

let farm_i (skl_gen, n) =
  if n = _0_ then invalid_arg "farm" "no worker" else
  let rec loop n p =
    if n = _1_ then skl_gen p else
    let q = Nat.div n _2_
    and r = Nat.rem n _2_ in
    Sk.c_par (loop q p) (loop (Nat.add q r) (Nat.add q p)) in
  loop n _0_
;;

let farm (skl, n) = farm_i ((fun _ -> skl), n);;

let farm_vector (skl, n) =
  let stream_of_vstream = get_exported of_sarray_gen in
  let vstream_of_stream = get_exported to_sarray_gen in
  stream_of_vstream ||| farm (skl, n) ||| vstream_of_stream
;;

let rails skla =
  let stream_of_vstream = get_exported of_sarray_gen in
  let vstream_of_stream = get_exported to_sarray_gen in
  stream_of_vstream ||| Sk.c_rails skla ||| vstream_of_stream
;;

(* Color handling *)

let colorize skl color =
  match skl with
  | Sk.Skl.Atom atom -> Sk.colorize atom color
  | Sk.Skl.Id _
  | Sk.Skl.Loop _
  | Sk.Skl.Par _
  | Sk.Skl.Pipe _
  | Sk.Skl.Prod _
  | Sk.Skl.Rails _
  | Sk.Skl.Sum _ -> skl  (* We cannot act on them *)
;;
