(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklstream.mli,v 1.46 2017/01/04 11:58:27 fclement Exp $ *)

(** {3 Stream model of the behavior of computational nodes}

 Computational nodes are a kind of light servers or processes or threads,
 that get there input from the wires on the network and send their results on
 the wires.

 Since there is no thunk that can be passed on the wires, there are no lazy
 values in the streams.
 Networks feature delay in values availability, hence there is a kind of lazy
 evaluation in networks that we modelize via lazy evaluation in the stream
 data structure.

 In networks, even to get the first element of a stream, we have to
 explicitely ask and wait; hence a stream is a lazy data structure that has
 to be forced to get its head or even to know that it is indeed empty.

 We have:
   [map_vector :
     ('a -> 'b) skel -> ('a array, 'b array) skel].
 Or equivalently:
   [map_vector :
     ('a stream -> 'b stream) -> 'a array stream -> 'b array stream].
 Given the following semantics:
   [map_vector skl \[| v1; ...; vn; |\] =
     \[| to_vector (skl (to_stream v1)); ...;
       to_vector (skl (to_stream vn)); |\]].
 To implement [map_vector], we build a farm that computes the resulting
 vector for each vector of the input stream, using the [skl] argument as
 worker.
 As a consequence, the workers of the same farm are used for each element
 of each vector; and since each vector is sent to the wires as a (sub) stream
 of vector elements, each farm worker must apply its skeleton to a sub stream
 and send the resulting sub stream to be collected and converted back into a
 vector that is the next element of the stream result.

 Hence we need:
 [vector_to_substream : 'a array -> 'a sub_stream],
 [substream_to_vector : 'a sub_stream -> 'a array],
 and [skl : 'a sub_stream -> 'b sub_stream].

 A computational node ends when it encounters the end of stream marker [Nil]
 (which is the end of lazy list).
 It also must propagate the end of stream marker along the wires before
 dying. *)

open Tools_common_types;;

(** {6 Stream definition} *)

type 'a t
and 'a stream = 'a t
(** The type of streams that run on the wires in parallel or are infinite
 data structures of lazy streams with ['a packet] elements. *)
;;

(** {6 Stream construction functions} *)

val singleton : 'a -> 'a t
(** [singleton x] returns the stream that has only [x] as element. *)
;;

val head : 'a t -> 'a
(** [head s] returns the first element of stream [s].

 Raise [End_of_stream], if [s] is empty. *)
;;

(** {6 Stream generation/consumption utilities} *)

val of_fun : (unit -> 'a option) -> 'a t
(** Generate a lazy stream from a function.
 The stream is closed with the value [None]. *)
;;

val to_fun : 'a t -> (unit -> 'a option)
(** Generate the function that successively gives the elements of a lazy
 stream. *)
;;

val of_list : 'a list -> 'a t
(** Generate a (finite) lazy stream from a list. *)
;;

val to_list : 'a t -> 'a list
(** Generate the list of the elements of a (finite) lazy stream.
 This operation completely unfreezes the stream argument. *)
;;

val of_vect : 'a array -> 'a t
(** Generate a (finite) lazy stream from a vector. *)
;;

val to_vect : 'a t -> 'a array
(** Generate the vector of the elements of a (finite) lazy stream.
 This operation completely unfreezes the stream argument. *)
;;

val of_channel : (in_channel -> 'a) -> in_channel -> 'a t
(** Generate a lazy stream from an input channel. *)
;;

val of_channel_by_lines : in_channel -> string t
(** Generate a lazy stream from the lines of an input channel. *)
;;

val of_channel_by_chars : in_channel -> char t
(** Generate a lazy stream from the characters of an input channel. *)
;;

(** {6 Operations on streams} *)

val join : 'a t * 'a t -> 'a t
(** Merges two lazy streams into a single stream with intermixed elements
 from both streams in a fair way (fairly peeks up alternatively elements from
 both streams). *)
;;

val fork : 'a t -> 'a t * 'a t
(** Splits a stream into two. *)
;;

val in_par : ('a t -> 'b t) * ('a t -> 'b t) -> 'a t -> 'b t;;

val split : ('a * 'b) t -> 'a t * 'b t;;

val combine : 'a t -> 'b t -> ('a * 'b) t;;

val in_prod : ('a t -> 'b t) * ('c t -> 'd t) -> ('a * 'c) t -> ('b * 'd) t;;

type ('a, 'b) sum =
  | Inl of 'a
  | Inr of 'b
(** The type of data handled by sum functions. *)
;;

type control = private
  | C_left
  | C_right
(** The private type for storing information used internally by sum handling
 functions. *)
;;

val sum_split : (('a, 'b) sum) t -> ('a t * 'b t) * control t
(** Splits an ['a + 'b] stream into two streams containing respectively
 elements of type ['a] and elements of type ['b].
 A control stream is generated, it is used by [sum_combine]. *)
;;

val sum_combine : ('a t * 'a t) * control t -> 'a t
(** Merges two streams of the same type using the control information stored
 in the control stream. *)
;;

val in_sum : ('a t -> 'c t) * ('b t -> 'c t) -> ('a, 'b) sum t -> 'c t;;

val fork_rails : size -> 'a t -> 'a t array
(** [fork_rails n s] splits stream [s] into an array of [n] streams. *)
;;

val join_rails : 'a t array -> 'a t
(** [join_rails sa] merges streams of array [sa] into one stream with
 intermixed elements. *)
;;

val on_rails : ('a t -> 'b t) array -> 'a t -> 'b t;;

val lift : ('a -> 'b) -> 'a t -> 'b t
(** [lift f] lifts regular function [f] into a function on streams. *)
;;

val unlift : ('a t -> 'b t) -> ('a -> 'b)
(** [unlift fs] turns a function on streams down to a regular function. *)
;;

val map : ('a -> 'b) -> 'a t -> 'b t
(** [map f s] maps function [f] on stream [s]. *)
;;

val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b t -> 'a t
(** Given a binary, associative, and commutative operator [op], its neutral
 element [zero] and a stream [s] = [x1; x2; ...; xn], then
 [fold_left op zero s] evaluates to
 [op (x1, op (x2, ..., (op (xn, zero))))]. *)
;;

val loop : ('a t -> bool t) -> ('a t -> 'a t) -> 'a t -> 'a t
(** Given [p], a predicate function on streams, and a regular function on
 streams [f], [loop p f s] returns the stream of fixpoints of [f] according
 to [p] using the elements of the input stream [s] as initialization. *)
;;

val iter : ('a -> unit) -> 'a t -> unit
(** Applies a function to each element of a (finite) lazy stream.
 This operation completely unfreezes the stream argument. *)
;;

(** {6 Serialization/deserialization for lists and vectors} *)

val serialize_stream_vector : 'a array t -> 'a t;;
val unserialize_stream_vector : 'a t -> 'a array t
(** The serialization/unserialization operations for vectors generalized to
 lazy streams of vectors. *)
;;

val of_sarray : 'a array t -> 'a t;;
val to_sarray : 'a t -> 'a array t
(** Synonyms of the two previous functions. *)
;;

(**/**)
(** {6 Undocumented features} *)

(** {6 IO functions} *)

val of_to_io_channel : IO.t -> 'a t -> 'b t;;
val to_io_channel : 'a t -> IO.oc -> unit;;
val of_io_channel : IO.ic -> 'a t;;
val stream_to_io_channel : 'a t -> IO.oc -> 'a t;;

(** {6 Debugging} *)

val print_stream : ('a, 'a t) Og_types.top_printer_lifter;;

val eq_stream : 'a t -> 'a t -> bool;;
