(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: batch.ml,v 1.41 2017/01/04 11:58:30 fclement Exp $ *)

(* Module used to spawn batch tasks *)

module Nat = Unatural;;
module L = Ulist;;
module U = Uunix;;

let _0_ = Nat._0_
and _255_ = Nat.unsafe_of_int 255
;;

let nat_unix_fork () =
  let pid = Unix.fork () in
  assert (pid >= 0);
  (* Now, we have pid >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  Nat.unsafe_of_int pid

and nat_unix_waitpid fl pid =
  let (pid, status) = Unix.waitpid fl (Nat.to_int pid) in
  assert (pid >= 0);
  (* Now, we have pid >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  (Nat.unsafe_of_int pid, status)

and nat_unix_kill pid = Unix.kill (Nat.to_int pid)

and nat_unix_alarm t = Unix.alarm (Nat.to_int t)
;;

open Tools_common_types;;

module type FIFO =
sig
  type 'a t;;

  val empty : 'a t;;
  val unshift : 'a t -> 'a -> 'a t;;
  val shift : 'a t -> 'a option * 'a t;;
end
;;

module Fifo : FIFO =
struct
  type 'a t =
    {
     inp : 'a list;
     outp : 'a list;
    }
  ;;

  let empty = { inp = []; outp = [] };;

  let unshift t x = { t with inp = x :: t.inp };;

  let shift t =
    let t = if t.outp = [] then { outp = L.rev t.inp; inp = []; } else t in
    match t.outp with
    | [] -> (None, t)
    | x :: xs -> (Some x, { t with outp = xs })
  ;;
end
;;

type shell_action =
  {
   program : string;
   args : string array;
  }
;;

type fail_action =
  | Fail_error of string
  | Fail_ignore
  | Fail_warn of string
;;

type task_type =
  | Background
  | Normal
  | Timeouted of timeout
;;

type task =
  {
   cmd : command; (* Command to launch to run the task *)
   on_fail : fail_action; (* What to do if it fails *)
  }

and command =
  | Batch of t
  | Control of (pid list -> t)
  | Echo of string
  | Fail
  | Pipe of t * (in_channel -> t)
  | Shell of shell_action * task_type

and t = task Fifo.t
and batch = t
;;

let empty = Fifo.empty;;

let add_shell_task batch fail cmd timeout =
  Fifo.unshift batch { cmd = Shell (cmd, timeout); on_fail = fail; }
;;

let add_subscript_task batch fail subbatch =
  Fifo.unshift batch { cmd = Batch subbatch; on_fail = fail; }
;;

let add_control_task batch fail c =
  Fifo.unshift batch { cmd = Control c; on_fail = fail; }
;;

let add_pipe_task batch fail subbatch f =
  Fifo.unshift batch { cmd = Pipe (subbatch, f); on_fail = fail; }
;;

let add_echo_task batch msg =
  Fifo.unshift batch { cmd = Echo msg; on_fail = Fail_ignore; }
;;

let add_fail_task batch msg =
  Fifo.unshift batch { cmd = Fail; on_fail = Fail_error msg; }
;;

type batch_error =
  | Batch_failure
  | Batch_shell_error of return_code * shell_action
  | Batch_timeout of shell_action
;;

type batch_run_return =
  | Sklml_Error of pid list * batch_error
  | Sklml_Ok of pid list
;;

let execvp_io p a oc =
  L.iter (Unix.dup2 oc) [ Unix.stdout; (* Unix.stderr; *) ];
  U.execvp p a
;;

let run_background_command oc cmd =
  let cpid = nat_unix_fork () in
  if cpid = _0_ then execvp_io cmd.program cmd.args oc else
  Sklml_Ok [ cpid; ]
;;

let run_normal_command oc cmd =
  let cpid = nat_unix_fork () in
  if cpid = _0_ then execvp_io cmd.program cmd.args oc else begin
    let (_, status) = nat_unix_waitpid [] cpid in
    match status with
    | Unix.WEXITED ret ->
      assert (ret >= 0);
      (* Now, we have ret >= 0.
         Therefore, the next Nat.unsafe_of_int will be safe. *)
      let ret = Nat.unsafe_of_int ret in
      if ret = _0_ then Sklml_Ok [] else
      Sklml_Error ([], Batch_shell_error (ret, cmd))
    | Unix.WSIGNALED _
    | Unix.WSTOPPED _ ->
      Sklml_Error ([], Batch_shell_error (_0_, cmd))
  end
;;

let run_timeouted_command timeout oc cmd =
  let cpid = nat_unix_fork () in
  if cpid = _0_ then
    let spawn { program = p; args = a; } =
      let cpid = nat_unix_fork () in
      let kill _ = nat_unix_kill cpid Sys.sigkill in
      if cpid = _0_ then execvp_io p a oc else begin
        Sys.set_signal Sys.sigalrm (Sys.Signal_handle kill);
        ignore (nat_unix_alarm timeout);
        let (_, status) =
          let rec loop () =
            try nat_unix_waitpid [] cpid with
            | Unix.Unix_error (Unix.EINTR, _, _) -> loop () in
          loop () in
        Sys.set_signal Sys.sigalrm Sys.Signal_default;
        let ret_code =
          match status with
          | Unix.WEXITED i -> i
          | Unix.WSIGNALED _ -> 255
          | Unix.WSTOPPED _ -> kill 0; 255 in
        exit ret_code
      end in
    spawn cmd
  else begin
    let (_, status) = nat_unix_waitpid [] cpid in
    match status with
    | Unix.WEXITED ret ->
      assert (ret >= 0);
      (* Now, we have ret >= 0.
         Therefore, the next Nat.unsafe_of_int will be safe. *)
      let ret = Nat.unsafe_of_int ret in
      if ret = _0_ then Sklml_Ok [] else
      if ret = _255_ then Sklml_Error ([], Batch_timeout cmd) else
      Sklml_Error ([], Batch_shell_error (ret, cmd))
    | Unix.WSIGNALED _
    | Unix.WSTOPPED _ ->
      Sklml_Error ([], Batch_shell_error (_0_, cmd))
  end
;;

let run_command t oc =
  let oc = Unix.descr_of_out_channel oc in
  match t with
  | Background -> run_background_command oc
  | Normal -> run_normal_command oc
  | Timeouted t -> run_timeouted_command t oc
;;

let print_error ec = Printf.fprintf ec "%s\n";;

let rec eval_batch b oc ec =
  let mk_pipe () =
    let (ic, oc) = Unix.pipe () in
    (Unix.in_channel_of_descr ic, Unix.out_channel_of_descr oc) in
  let eval_pipe b f =
    let (ic, oc_pipe) = mk_pipe () in
    match eval_batch b oc_pipe ec with
    | Sklml_Ok l ->
      begin match eval_batch (f ic) oc ec with
      | Sklml_Ok l' -> Sklml_Ok (l @ l')
      | Sklml_Error (l', e) -> Sklml_Error (l @ l', e)
      end
    | Sklml_Error _ as err -> err in
  let eval_command cmd pids =
    match cmd with
    | Batch b -> eval_batch b oc ec
    | Control f -> eval_batch (f pids) oc ec
    | Echo m -> print_error ec m; Sklml_Ok []
    | Fail -> Sklml_Error ([], Batch_failure)
    | Pipe (b, f) -> eval_pipe b f
    | Shell (cmd, t) -> run_command t oc cmd in
  let rec loop b pids =
    match Fifo.shift b with
    | (None, _) -> Sklml_Ok pids
    | (Some { cmd = cmd; on_fail = f; }, b) ->
      begin match eval_command cmd pids with
      | Sklml_Ok l -> loop b (l @ pids)
      | Sklml_Error (l, e) ->
        begin match f with
        | Fail_error em -> print_error ec em; Sklml_Error (l @ pids, e)
        | Fail_warn em -> print_error ec em; loop b pids
        | Fail_ignore -> loop b pids
        end
      end in
  loop b []
;;
