(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: node.ml,v 1.34 2017/02/05 11:42:03 fclement Exp $ *)

module Nat = Unatural;;
module Perv = Upervasives;;
module L = Ulist;;
module L1 = Ulist1;;
module S = Ustring;;
module A = Uarray;;
module Ar = Uaref;;

let _0_ = Nat._0_;;

let string_concat_array a = S.concat " " (A.to_list a);;

open Tools_common_types;;

let { tagged_report_action = report_action;
      tagged_debug = debug;
      tagged_debug_action = debug_action; _; } = Options.services
;;

let debug fname = debug (Printf.sprintf "Node.%s" fname);;

type kind =
  | Master
  | Slave
;;

type node =
  {
   name : string;
   host : string;
   shell : string;
   color : color;
   daemon_path : string;
   program_path : string option;
  }
;;

type master_node = node
and slave_node = node
;;

let default_master_node =
  {
   name = "Master";
   host = "127.0.0.1";
   shell = "sh -c '%c'";
   color = 2;
   daemon_path = "sklmlpard";
   program_path = None;
  }
;;

let { Ar.get = get_master_node; set = set_master_node; } =
  Ar.make_aref_single "master node" default_master_node
;;

let default_slave_node () =
  { (get_master_node ()) with name = "Slave node #0"; }
;;

let add h (k, v) =
  if Hashtbl.mem h k then Error.warn_duplicated_entry k else
  Hashtbl.add h k v
;;

let find_opt h k =
  try Some (Hashtbl.find h k) with
  | Not_found -> None
;;

let get_int h k d =
  match find_opt h k with
  | None -> d
  | Some (Parsers.Nat i) -> Nat.to_int i
  | Some (Parsers.Str _) -> Error.warn_expected_int k; d

and gen_get_str f h k d =
  match find_opt h k with
  | None -> d
  | Some (Parsers.Nat _) -> Error.warn_expected_string k; d
  | Some (Parsers.Str s) -> f s
;;

let get_str = gen_get_str Perv.id
and get_str_opt = gen_get_str (fun s -> Some s)
;;

let mk_node kind { Options.daemon_path; _; } default node_settings =
  let debug = debug "mk_node" in
  let h = Hashtbl.create 10 in
  L.iter (add h) node_settings;
  let name = get_str h "name" default.name
  and host = get_str h "host" default.host
  and shell = get_str h "shell" default.shell
  and color = get_int h "color" default.color
  and daemon_path =
    match daemon_path with
    | None -> get_str h "daemon_path" default.daemon_path
    | Some dp -> dp
  and program_path =
    let k = "program_path" in
    let pp_opt = get_str_opt h k default.program_path in
    match (kind, pp_opt) with
    | (Master, Some _) -> Error.warn_meaningless k; None
    | (Slave, Some _)
    | (_, None) -> pp_opt in
  debug (Printf.sprintf "name = %S" name);
  debug (Printf.sprintf "host = %S" host);
  debug (Printf.sprintf "shell = %S" shell);
  debug (Printf.sprintf "color = %d" color);
  debug (Printf.sprintf "daemon_path = %S" daemon_path);
  debug
    (Printf.sprintf "program_path = %S"
      (match program_path with None -> "[None]" | Some pp -> pp));
  {
    name = name;
    host = host;
    shell = shell;
    color = color;
    daemon_path = daemon_path;
    program_path = program_path;
  }
;;

let mk_master options settings =
  debug "mk_master" "building node";
  let node = mk_node Master options default_master_node settings in
  set_master_node node;
  node

and mk_slave options settings =
  debug "mk_slave" "building node";
  mk_node Slave options (default_slave_node ()) settings
;;

let default_d_fmt = format_of_string "";;

let { Ar.get = get_d_fmt; set = set_d_fmt; } = Ar.make_aref default_d_fmt;;

let report_d_fmt = format_of_string " -r"
and debug_d_fmt = format_of_string " -d"
;;

let action_report_d_fmt () =
  set_d_fmt report_d_fmt;
  "transmitting report mode to Sklml parallel daemon"
and action_debug_d_fmt () =
  set_d_fmt debug_d_fmt;
  "transmitting debug mode to Sklml parallel daemon"
;;

(* Build the batch launching the master Sklml parallel daemon program. *)
let compile_master
    options nslaves { name; host; shell; daemon_path = dp; _; } =
  let debug = debug "compile_master" in
  report_action action_report_d_fmt;
  debug_action "Node.compile_master" action_debug_d_fmt;
  let d_fmt =
    format_of_string "%S -mode master -slave-number %d" ^^ get_d_fmt () in
  let d_cmd = Printf.sprintf d_fmt dp (Nat.to_int nslaves) in
  debug (Printf.sprintf "daemon command = %S" d_cmd);
  let sh_cmd = Parsers.expand_shell_command options host d_cmd shell in
  debug (Printf.sprintf "shell command = %S" (string_concat_array sh_cmd));
  if A.length sh_cmd = _0_ then Batch.empty else
  (* Now, we have A.length sh_cmd > 0.
     Therefore, the next A.unsafe_get will be safe. *)
  Batch.add_shell_task
    Batch.empty
    (Error.batch_fail_start_master name)
    { Batch.program = A.unsafe_get sh_cmd _0_; Batch.args = sh_cmd; }
    Batch.Background
;;

type addr = string;;

let mk_addr node port = Printf.sprintf "%s:%d" node.host (Nat.to_int port);;

(* FIXME: how arguments of program_path are handled?
let local_program_command_line
    { Options.sklml_program_command_line = pcl; _; } = function
  | None -> A.unsafe_get pcl _0_
  | Some pp -> pp
;;
*)

let local_program_command_line
    { Options.sklml_program_command_line = pcl; _; } program_path =
  (* A.length pcl > 0, since it is built from a nonempty list.
     Therefore, the next A.unsafe_set will be safe. *)
  begin match program_path with
  | None -> ()
  | Some pp -> A.unsafe_set pcl _0_ pp
  end;
  string_concat_array pcl
;;

(* Build the batches that will run the commands to launch the remote slave
   Sklml parallel launcher daemon programs. *)
let compile_slaves options master_node port slave_nodes =
  let debug = debug "compile_slaves" in
  let master_addr = mk_addr master_node port in
  debug (Printf.sprintf "master_addr = %S" master_addr);
  let compile b { name; host; shell; color; daemon_path; program_path; } =
    let pcl = local_program_command_line options program_path in
    report_action action_report_d_fmt;
    debug_action "Node.compile_slave" action_debug_d_fmt;
    let d_fmt =
      format_of_string
        "%S -mode slave \
         -master-address %s -sklml-program-path %S -color %d" ^^
      get_d_fmt () in
    let d_cmd = Printf.sprintf d_fmt daemon_path master_addr pcl color in
    debug (Printf.sprintf "daemon command = %S" d_cmd);
    let sh_cmd = Parsers.expand_shell_command options host d_cmd shell in
    debug (Printf.sprintf "shell command = %S" (string_concat_array sh_cmd));
    if A.length sh_cmd = _0_ then b else
    (* Now, we have A.length sh_cmd > 0.
       Therefore, the next A.unsafe_get will be safe. *)
    Batch.add_shell_task
      b
      (Error.batch_fail_launch_slave name)
      { Batch.program = A.unsafe_get sh_cmd _0_; Batch.args = sh_cmd; }
      Batch.Background in
  L1.fold_left compile Batch.empty slave_nodes
;;
