(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.33 2017/05/15 06:54:55 fclement Exp $ *)

module Nat = Unatural;;
module Perv = Upervasives;;
module L = Ulist;;
module S = Ustring;;
module A = Uarray;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module Pic = Upph_id_card;;

let _0_ = Nat._0_
and _10_ = Nat.unsafe_of_int 10
and _100_ = Nat.unsafe_of_int 100
;;

let nat_uaref_make_aref mk_aref_int name init =
  let { Ar.get; set; } = mk_aref_int name (Nat.to_int init) in
  (* We will always have get () > 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  let get () = Nat.unsafe_of_int (get ()) in
  let set n = set (Nat.to_int n) in
  { Ar.get; set; }
;;

let a_find_index p a =
  let len = A.length a in
  let rec loop i =
    if i >= len then raise Not_found else
    (* Now, we have 0 <= i < len = length a.
       Therefore, the following A.unsafe_get will be safe. *)
    if p (A.unsafe_get a i) then i else loop (Nat.succ i) in
  loop _0_
;;

let qual = Perv.qualify "Options";;
let invalid_arg = Perv.invalid_arg_qual qual;;

let nat_set set n =
  if n < 0 then invalid_arg "nat_set" "negative number" else
  (* Now, we have n >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  set (Nat.unsafe_of_int n)
;;

module Opts = Tools_options;;

open Tools_common_types;;

(* Command-line options. *)

let sklml_program_command_line_name = "sklml program command line"
and configuration_filename_name = "configuration filename"
and maximum_error_message_length_name = "maximum error message length"
and daemon_path_name = "daemon path"
and timeout_name = "timeout"
;;

let init_sklml_program_command_line = []
and default_configuration_filename = "-"
and default_maximum_error_message_length = _100_
and default_daemon_path = None
and default_timeout = _10_
;;

let add_sklml_program_command_line argv add_item =
  try
    let i = a_find_index (fun arg -> arg = "--") argv in
    let add_argv_item i =
      (* add_argv_item is always called with 0 <= i < A.length a - 1.
         Therefore, the following A.unsafe_get will be safe. *)
      add_item (A.unsafe_get argv i) in
    Nat.for_loop (Nat.succ i) (Nat.pred (A.length argv)) add_argv_item
  with
  | Not_found -> ()
;;

(* Function set_sklml_program_command_line calls
   add_sklml_program_command_line to iterate add_item to all components of
   Sys.argv after the first ``--'' component that starts the Sklml program
   name and its arguments. *)
let (get_sklml_program_command_line, set_sklml_program_command_line) =
  let { Ar.get_list; add_item; _; } =
 Ar.make_aref_nonempty_list
      sklml_program_command_line_name init_sklml_program_command_line in
  let get_array () = A.of_list (get_list ()) in
  (* We don't mind the string argument passed to function
     set_sklml_program_command_line: this string is just the rest of the
     command line of the main program, but we need the exact shell tokenized
     list of arguments for the Sklml program. *)
  let set_array _ = add_sklml_program_command_line Sys.argv add_item in
  (get_array, set_array)

and { Ar.get = get_configuration_filename;
      set = set_configuration_filename; } =
  Ar.make_aref_single
    configuration_filename_name default_configuration_filename

and { Ar.get = get_maximum_error_message_length;
      set = set_maximum_error_message_length; } =
  nat_uaref_make_aref Ar.make_aref_single_positive_int
    maximum_error_message_length_name default_maximum_error_message_length

and (get_daemon_path, set_daemon_path) =
  let { Ar.get; set; } =
    Ar.make_aref_single daemon_path_name default_daemon_path in
  let set s = set (Some s) in
  (get, set)

and { Ar.get = get_timeout; set = set_timeout; } =
  nat_uaref_make_aref Ar.make_aref_single_positive_int
  timeout_name default_timeout
;;

let { Ok.long = program_command_line_lok;
      short = program_command_line_sok_st; } =
  Ok.make_option_keys
    sklml_program_command_line_name
    (Ok.Short (Ok.of_string "--"))

and { Ok.long = configuration_filename_lok;
      short = configuration_filename_sok_st; } =
  Ok.make_option_keys
    configuration_filename_name
    (Ok.Short (Ok.of_string "-conf"))

and { Ok.long = maximum_error_message_length_lok;
      short = maximum_error_message_length_sok_st; } =
  Ok.make_option_keys
    maximum_error_message_length_name
    (Ok.Short (Ok.of_string "-err-len"))

and { Ok.long = daemon_path_lok; short = daemon_path_sok_st; } =
  Ok.make_option_keys daemon_path_name Ok.Automatic

and { Ok.long = timeout_lok; short = timeout_sok_st; } =
  Ok.make_option_keys timeout_name Ok.Automatic
;;

let program_command_line_rco = {
  Co.key = program_command_line_lok;
  spec = Arg.Rest set_sklml_program_command_line;
  doc =
    Printf.sprintf
      "<command_line>\n\
       \tTreat the remaining arguments as the Sklml program command line\n\
       \tto run in parallel.\n\
       \tThese must be the last arguments of the present command.\n\
       \tThe command name (first element of <command_line>) can be locally\n\
       \tmodified in the configuration file.";
}

and configuration_filename_rco = {
  Co.key = configuration_filename_lok;
  spec = Arg.String set_configuration_filename;
  doc =
    Printf.sprintf
      "<path>\n\
       \tSet the configuration filename to <path>.\n\
       \tThe special filename \"-\" stands for the standard input.\n\
       \tDefault is %S"
      (get_configuration_filename ());
}

and maximum_error_message_length_rco = {
  Co.key = maximum_error_message_length_lok;
  spec = Arg.Int (nat_set set_maximum_error_message_length);
  doc =
    Printf.sprintf
      "<nat>\n\
       \tSet the maximum length of error messages to <nat>.\n\
       \tThis is specific to the parsing of the configuration file.\n\
       \tIt is a positive integer. Default is %d."
      (Nat.to_int (get_maximum_error_message_length ()));
}

and daemon_path_rco = {
  Co.key = daemon_path_lok;
  spec = Arg.String set_daemon_path;
  doc =
    Printf.sprintf
      "<path>\n\
       \tSet the path of the Sklml parallel launcher daemon to <path>.\n\
       \tOverrides settings from the configuration file.";
}

and timeout_rco = {
  Co.key = timeout_lok;
  spec = Arg.Int (nat_set set_timeout);
  doc =
    Printf.sprintf
      "<nat>\n\
       \tSet the maximum amount of time in seconds waited before deciding\n\
       \tthat the Sklml computation network is down.\n\
       \tIt is a positive integer. Default is %d."
      (Nat.to_int (get_timeout ()));
}
;;

let command_kind = "parallel launcher"
and command_suffix = "run"
;;

let rcos_with_sok = [
  (program_command_line_sok_st, program_command_line_rco);
  (configuration_filename_sok_st, configuration_filename_rco);
  (maximum_error_message_length_sok_st, maximum_error_message_length_rco);
  (daemon_path_sok_st, daemon_path_rco);
  (timeout_sok_st, timeout_rco);
]
;;

let doc_mandatory_arguments = "-- <command line>";;

let project_name = Sklml_version.pph_id_card.Pic.project_name

let sklml_command_kind = Printf.sprintf "%s %s" project_name command_kind

let command_title = Printf.sprintf "%s%s" project_name command_suffix;;

let add_options () =
  Co.application_add_all_rco_with_sk rcos_with_sok
;;

add_options ();;

Pic.add_app_version_options Sklml_version.pph_id_card None;;

(* Sys.argv is never empty.
   Therefore, the next A.unsafe_get will be safe. *)
let cmd_name = A.unsafe_get Sys.argv _0_;;
Co.set_usage_msg
  (Format.sprintf
     "@[<v>\
      @ \
      %s -- %s@ \
      @ \
      Usage:@ \
      @ \
      %s [<option>...] %s@ \
      @ \
      Arguments are:\
      @]@."
     cmd_name sklml_command_kind
     cmd_name doc_mandatory_arguments)
;;

(* Report/debug flags and warning/error messages. *)

let { Tools_common_types.tagged_error = error; _; } as services =
  Opts.add_tagged_services command_title
;;

(* Parsing the comand line and the configuration file. *)

type t =
  {
   maximum_error_message_length : length;
   daemon_path : string option;
   timeout : timeout;
   sklml_program_command_line : string array;
  }
and options = t
;;

type sections = string;;

let get_options () = {
  maximum_error_message_length = get_maximum_error_message_length ();
  daemon_path = get_daemon_path ();
  timeout = get_timeout ();
  sklml_program_command_line = get_sklml_program_command_line ();
}
;;

let get_sections () =
  let get_conf () =
    let fname = get_configuration_filename () in
    let ic = if fname = "-" then stdin else open_in fname in
    let rec loop accu =
      try loop (input_line ic :: accu) with
      | End_of_file -> L.rev accu in
    S.concat "\n" (loop []) in
  try get_conf () with
  | Sys_error msg ->
    error (Printf.sprintf "Cannot retrieve configuration: %s\n" msg)
;;

let parse () =
  Co.parse_all_services ();
  Co.parse ();
  let options = get_options () in
  let sections = get_sections () in
  (options, sections)
;;
