(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parser.ml,v 1.20 2017/01/04 11:58:30 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module S = Ustring;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

type 'a t = { parse : string -> ('a * string) option; }
and 'a parser = 'a t
;;

let parse p s =
  match p.parse s with
  | None -> None
  | Some (x, _) -> Some x
;;

let full_parse p = p.parse;;

let return x = { parse = fun s -> Some (x, s) };;
let zero = { parse = fun _ -> None };;
let bind p f =
  let parse s =
    match p.parse s with
    | None -> None
    | Some (x, s') -> (f x).parse s' in
  { parse; }
;;

let ( >>= ) = bind;;
let ( >> ) p0 p1 = p0 >>= (fun _ -> p1);;

let plus p0 p1 =
  let parse s =
    let ret = p0.parse s in
    if ret <> None then ret else
    p1.parse s in
  { parse; }
;;

let ( +++ ) = plus;;

let char c =
  let parse s =
    let len = S.length s in
    if len = _0_ then None else
    (* Now, we have len > 0.
       Therefore, the next S.unsafe_get will be safe and the next Nat.pred
       will not fail. *)
    if S.unsafe_get s _0_ <> c then None else
    Some (c, S.sub s _1_ (Nat.pred len)) in
  { parse; }
;;

let eof =
  { parse = fun s -> if S.length s = _0_ then Some ((), "") else None; }
;;

let rec any = function
  | [] -> zero
  | p :: ps -> p +++ (any ps)
;;

let many p =
  let rec loop acc =
    (p >>= (fun c -> loop (c :: acc))) +++
    (return (L.rev acc)) in
  loop []
;;

let many1 p = p >>= (fun x -> many p >>= (fun l -> return (x :: l)));;

let many_till pcond p =
  let rec loop acc =
    (pcond >> return (L.rev acc)) +++
    (p >>= fun x -> loop (x :: acc)) in
  loop []
;;

let maybe p = (p >>= (fun x -> return (Some x))) +++ (return None);;

let times n p =
  let rec loop n acc =
    if n = _0_ then return (L.rev acc) else
    (* Now, we have n > 0.
       Therefore, the next Nat.pred will not fail. *)
    p >>= (fun x -> loop (Nat.pred n) (x :: acc)) in
  loop n []
;;

let less_than n p =
  let rec loop n acc =
    if n = _0_ then return (L.rev acc) else
    (* Now, we have n > 0.
       Therefore, the next Nat.pred will not fail. *)
    (p >>= (fun x -> loop (Nat.pred n) (x :: acc))) +++
    return (L.rev acc) in
  loop n []
;;
